from .module import Module
from .. import functional as F
from ...tensor import Tensor


class Sigmoid(Module):
    def forward(self, x) -> Tensor:
        return F.sigmoid(x)

    def __repr__(self) -> str:
        return "{}()".format(self.__class__.__name__)


class Tanh(Module):
    def forward(self, x) -> Tensor:
        return F.tanh(x)

    def __repr__(self) -> str:
        return "{}()".format(self.__class__.__name__)


class ReLU(Module):
    def forward(self, x) -> Tensor:
        return F.relu(x)

    def __repr__(self) -> str:
        return "{}()".format(self.__class__.__name__)


class LeakyReLU(Module):
    def __init__(self, alpha: float) -> None:
        super().__init__()
        self.alpha = float(alpha)

    def forward(self, x) -> Tensor:
        return F.leaky_relu(x, self.alpha)

    def __repr__(self) -> str:
        return "{}(alpha={})".format(self.__class__.__name__, self.alpha)


class Softmax(Module):
    def __init__(self, axis=None) -> None:
        super().__init__()
        self.axis = axis

    def forward(self, x) -> Tensor:
        return F.softmax(x, self.axis)

    def __repr__(self) -> str:
        return "{}(axis={})".format(self.__class__.__name__, self.axis)


import numpy as np

from .. import tensor


def linear(x: tensor.Tensor, weight: tensor.Tensor, bias: tensor.Tensor):
    affine = x @ weight
    if bias is not None:
        affine = affine + bias
    return affine


class sigmoid(tensor.UnaryOperator):
    '''Sigmoid运算，我们前向传播避免了溢出问题'''
    def forward(self, x: tensor.Tensor) -> np.ndarray:
        sigmoid = np.zeros(x.shape)
        sigmoid[x.data > 0] = 1 / (1 + np.exp(-x.data[x.data > 0]))
        sigmoid[x.data <= 0] = 1 - 1 / (1 + np.exp(x.data[x.data <= 0]))
        return sigmoid

    def grad_fn(self, x: tensor.Tensor, grad: np.ndarray) -> np.ndarray:
        return self.data * (1 - self.data) * grad


class tanh(tensor.UnaryOperator):
    '''Tanh运算，我们前向传播避免了溢出问题'''
    def forward(self, x: tensor.Tensor) -> np.ndarray:
        tanh = np.zeros(x.shape)
        tanh[x.data > 0] = 2 / (1 + np.exp(-2 * x.data[x.data > 0])) - 1
        tanh[x.data <= 0] = 1 - 2 / (1 + np.exp(2 * x.data[x.data <= 0]))
        return tanh

    def grad_fn(self, x: tensor.Tensor, grad: np.ndarray) -> np.ndarray:
        return (1 - self.data**2) * grad


def relu(x: tensor.Tensor):
    return tensor.maximum(0., x)


def leaky_relu(x: tensor.Tensor, alpha: float):
    return tensor.maximum(x, alpha * x)


def softmax(x: tensor.Tensor, axis=None, keepdims=False):
    '''Softmax函数'''
    x_sub_max = x - np.max(x.data)
    exp_ = tensor.exp(x_sub_max)
    return exp_ / tensor.sum(exp_, axis=axis, keepdims=keepdims)


def log_softmax(x: tensor.Tensor, axis=None, keepdims=False):
    '''log-softmax函数'''
    x_sub_max = x - np.max(x.data)
    return x_sub_max - tensor.log(
        tensor.sum(tensor.exp(x_sub_max), axis=axis, keepdims=keepdims))


class __im2col1d(tensor.UnaryOperator):
    def __init__(self, x: tensor.Tensor, kernel_size: int,
                 stride: int) -> None:
        self.N, self.in_channels, self.n_features = x.shape
        self.kernel_size = kernel_size
        self.stride = stride
        self.n_output = (self.n_features - self.kernel_size) // stride + 1
        super().__init__(x)

    def forward(self, x: tensor.Tensor) -> np.ndarray:
        col = np.zeros(
            (self.N, self.in_channels, self.n_output, self.kernel_size))

        for i in range(self.kernel_size):
            i_max = i + self.n_output * self.stride
            col[..., i] = x.data[..., i:i_max:self.stride]

        return col

    def grad_fn(self, x: tensor.Tensor, grad: np.ndarray) -> np.ndarray:
        grad_x = np.zeros((self.N, self.in_channels, self.n_features))
        for i in range(self.kernel_size):
            i_max = i + self.n_output * self.stride
            grad_x[..., i:i_max:self.stride] += grad[..., i]

        return grad_x


class __pad1d(tensor.UnaryOperator):
    def __init__(self, x: tensor.Tensor, pad_width=0) -> None:
        self.pad_width = pad_width
        super().__init__(x)

    def forward(self, x: tensor.Tensor) -> np.ndarray:
        return np.pad(x.data, [(0, 0), (0, 0),
                               (self.pad_width, self.pad_width)], 'constant')

    def grad_fn(self, x: tensor.Tensor, grad: np.ndarray) -> np.ndarray:
        if self.pad_width == 0:
            return grad[...]
        return grad[..., self.pad_width:-self.pad_width]


def conv1d(x: tensor.Tensor,
           kernel: tensor.Tensor,
           padding: int = 0,
           stride: int = 1):
    '''一维卷积函数

    基于im2col实现的一维卷积.
    
    Parameters
    ----------
    x : Tensor
        输入数据，形状为(N, in_channels, n_features);
    kernel : Tensor
        卷积核，形状为(out_channels, in_channels, kernel_size);
    padding : int, default=0
        对输入特征两边补0数量;
    stride : int, default=1
        卷积步长.
    '''
    kernel_size = kernel.shape[-1]
    pad_x = __pad1d(x, padding)
    col = __im2col1d(pad_x, kernel_size, stride)
    return (col @ kernel.transpose(1, 2, 0)).sum(1).transpose(0, 2, 1)


def max_pool1d(x: tensor.Tensor, kernel_size: int, stride: int, padding=0):
    '''一维池化函数

    基于im2col实现的一维池化.`
    
    Parameters
    ----------
    x : Tensor
        输入数据，形状为(N, in_channels, n_features);
    kernel_size : int
        池化核大小;
    stride : int
        卷积步长;
    padding : int, default=0
        对输入特征两边补0数量.
    '''
    pad_x = __pad1d(x, padding)
    col = __im2col1d(pad_x, kernel_size, stride)
    return col.max(-1)


def avg_pool1d(x: tensor.Tensor, kernel_size: int, stride: int, padding=0):
    '''一维平均池化函数

    基于im2col实现的一维池化.`
    
    Parameters
    ----------
    x : Tensor
        输入数据，形状为(N, in_channels, n_features);
    kernel_size : int
        池化核大小;
    stride : int
        卷积步长;
    padding : int, default=0
        对输入特征两边补0数量.
    '''
    pad_x = __pad1d(x, padding)
    col = __im2col1d(pad_x, kernel_size, stride)
    return col.mean(-1)


class __im2col2d(tensor.UnaryOperator):
    def __init__(self, x: tensor.Tensor, kernel_size, stride: int) -> None:
        self.N, self.in_channels, self.n_h, self.n_w = x.shape
        self.kernel_size = kernel_size
        self.stride = stride
        self.out_h, self.out_w = (
            self.n_h - self.kernel_size) // self.stride + 1, (
                self.n_w - self.kernel_size) // self.stride + 1
        super().__init__(x)

    def forward(self, x: tensor.Tensor) -> np.ndarray:
        col = np.zeros((self.N, self.in_channels, self.kernel_size,
                        self.kernel_size, self.out_h, self.out_w))
        for i in range(self.kernel_size):
            i_max = i + self.out_h * self.stride
            for j in range(self.kernel_size):
                j_max = j + self.out_w * self.stride
                col[:, :, i, j, :, :] = x.data[:, :, i:i_max:self.stride,
                                               j:j_max:self.stride]

        return col

    def grad_fn(self, x: tensor.Tensor, grad: np.ndarray) -> np.ndarray:
        grad_col = grad
        grad_x = np.zeros((self.N, self.in_channels, self.n_h, self.n_w))
        for i in range(self.kernel_size):
            i_max = i + self.out_h * self.stride
            for j in range(self.kernel_size):
                j_max = j + self.out_w * self.stride
                grad_x[:, :, i:i_max:self.stride,
                       j:j_max:self.stride] = grad_col[:, :, i, j, :, :]
        return grad_x


class __pad2d(tensor.UnaryOperator):
    def __init__(self, x: tensor.Tensor, pad_width=0) -> None:
        self.pad_width = pad_width
        super().__init__(x)

    def forward(self, x: tensor.Tensor) -> np.ndarray:
        return np.pad(x.data, [(0, 0), (0, 0),
                               (self.pad_width, self.pad_width),
                               (self.pad_width, self.pad_width)], 'constant')

    def grad_fn(self, x: tensor.Tensor, grad: np.ndarray) -> np.ndarray:
        if self.pad_width == 0:
            return grad[...]
        return grad[..., self.pad_width:-self.pad_width,
                    self.pad_width:-self.pad_width]


def conv2d(x: tensor.Tensor,
           kernel: tensor.Tensor,
           padding: int = 0,
           stride: int = 1):
    '''二维卷积函数

    基于im2col实现的二维卷积. 为了实现上的方便，我们不考虑长宽不同的卷积核，步长和补零。
    
    Parameters
    ----------
    x : Tensor
        输入数据，形状为(N, in_channels, n_height, n_width);
    kernel : Tensor
        卷积核，形状为(out_channels, in_channels, kernel_height, kernel_width);
    padding : int, default=0
        对输入图片周围补0数量;
    stride : int, default=1
        卷积步长.
    '''
    N, _, _, _ = x.shape
    out_channels, _, kernel_size, _ = kernel.shape
    pad_x = __pad2d(x, padding)
    col = __im2col2d(pad_x, kernel_size, stride)
    out_h, out_w = col.shape[-2:]
    col = col.transpose(0, 4, 5, 1, 2, 3).reshape(N * out_h * out_w, -1)
    col_filter = kernel.reshape(out_channels, -1).T
    out = col @ col_filter
    return out.reshape(N, out_h, out_w, -1).transpose(0, 3, 1, 2)


def max_pool2d(x: tensor.Tensor, kernel_size: int, stride: int, padding=0):
    '''二维卷积函数池化

    基于im2col实现的二维卷积. 为了实现上的方便，我们不考虑长宽不同的kernel_size，步长和补零。
    
    Parameters
    ----------
    x : Tensor
        输入数据，形状为(N, in_channels, n_height, n_width);
    kernel_size : int
        池化核尺寸;
    stride : int, default=1
        卷积步长;
    padding : int, default=0
        对输入图片周围补0数量;
    '''
    N, in_channels, _, _ = x.shape
    pad_x = __pad2d(x, padding)
    col = __im2col2d(pad_x, kernel_size, stride)
    out_h, out_w = col.shape[-2:]
    col = col.transpose(0, 4, 5, 1, 2, 3).reshape(
        -1,
        kernel_size * kernel_size,
    )
    out = col.max(1)
    out = out.reshape(N, out_h, out_w, in_channels).transpose(0, 3, 1, 2)
    return out


def avg_pool2d(x: tensor.Tensor, kernel_size: int, stride: int, padding=0):
    '''二维平均池化

    基于im2col实现的二维池化. 为了实现上的方便，我们不考虑长宽不同的kernel_size，步长和补零。
    
    Parameters
    ----------
    x : Tensor
        输入数据，形状为(N, in_channels, n_height, n_width);
    kernel_size : int
        池化核尺寸;
    stride : int, default=1
        卷积步长;
    padding : int, default=0
        对输入图片周围补0数量;
    '''
    N, in_channels, _, _ = x.shape
    pad_x = __pad2d(x, padding)
    col = __im2col2d(pad_x, kernel_size, stride)
    out_h, out_w = col.shape[-2:]
    col = col.transpose(0, 4, 5, 1, 2, 3).reshape(
        -1,
        kernel_size * kernel_size,
    )
    out = col.mean(1)
    out = out.reshape(N, out_h, out_w, in_channels).transpose(0, 3, 1, 2)
    return out


def mse_loss(y_pred, y_true, reduction='mean'):
    '''均方误差'''
    square_sum = tensor.square(y_pred - y_true)
    if reduction == 'mean':
        return tensor.mean(square_sum)
    elif reduction == 'sum':
        return tensor.sum(square_sum)
    else:
        assert 0, "reduction must be mean or sum."


def nll_loss(y_pred, y_true, reduction='mean'):
    '''负对数似然'''
    nll = -y_pred * y_true
    if reduction == 'mean':
        return tensor.mean(nll)
    elif reduction == 'sum':
        return tensor.sum(nll)
    else:
        assert 0, "reduction must be mean or sum."


def cross_entropy_loss(y_pred, y_true, reduction='mean'):
    '''交叉熵损失'''
    update_y_pred = y_pred - np.max(y_pred.data)
    log_sum_exp = tensor.log(
        tensor.sum(tensor.exp(update_y_pred), 1, keepdims=True))
    nll = (log_sum_exp - update_y_pred) * y_true
    if reduction == 'mean':
        return tensor.mean(nll)
    elif reduction == 'sum':
        return tensor.sum(nll)
    else:
        assert 0, "reduction must be mean or sum."



from .module import Module
from ..parameter import Parameter
from ... import tensor


class BatchNorm1d(Module):
    def __init__(self,
                 num_features: int,
                 eps: float = 1e-5,
                 momentum: float = 0.1) -> None:
        super().__init__()
        self.num_features = num_features
        self.eps = eps
        self.momentum = momentum
        self.running_mean = tensor.zeros(self.num_features)
        self.running_var = tensor.zeros(self.num_features)
        self.scale = Parameter(tensor.randn(self.num_features))
        self.shift = Parameter(tensor.zeros(self.num_features))

    def forward(self, x: tensor):
        if self._train:
            mean = x.mean(0)
            center_data = x - mean
            var = tensor.mean(tensor.square(center_data), 0)
            std_data = center_data / tensor.sqrt(var + self.eps)

            self.running_mean *= (1 - self.momentum)
            self.running_mean += self.momentum * mean
            self.running_var *= (1 - self.momentum)
            self.running_var += self.momentum * var

            return std_data * self.scale + self.shift
        else:
            return (x - self.running_mean) * self.scale / tensor.sqrt(
                self.running_var + self.eps) + self.shift

    def __repr__(self) -> str:
        return "{}(num_features={}, momentum={})".format(
            self.__class__.__name__,
            self.num_features,
            self.momentum,
        )


class BatchNorm2d(Module):
    def __init__(self,
                 num_features: int,
                 eps: float = 1e-5,
                 momentum: float = 0.1) -> None:
        super().__init__()
        self.num_features = num_features
        self.eps = eps
        self.momentum = momentum
        self.running_mean = tensor.zeros((1, self.num_features, 1, 1))
        self.running_var = tensor.zeros((1, self.num_features, 1, 1))
        self.scale = Parameter(tensor.randn(1, self.num_features, 1, 1))
        self.shift = Parameter(tensor.zeros((1, self.num_features, 1, 1)))

    def forward(self, x: tensor):
        if self._train:
            mean = x.mean((0, 2, 3), keepdims=True)
            center_data = x - mean
            var = tensor.mean(tensor.square(center_data), (0, 2, 3),
                              keepdims=True)
            std_data = center_data / tensor.sqrt(var + self.eps)

            self.running_mean *= (1 - self.momentum)
            self.running_mean += self.momentum * mean
            self.running_var *= (1 - self.momentum)
            self.running_var += self.momentum * var

            return std_data * self.scale + self.shift
        else:
            return (x - self.running_mean) * self.scale / tensor.sqrt(
                self.running_var + self.eps) + self.shift

    def __repr__(self) -> str:
        return "{}(num_features={}, momentum={})".format(
            self.__class__.__name__,
            self.num_features,
            self.momentum,
        )


from .module import Module
from ..parameter import Parameter
from .. import init
from .. import functional as F
from ...tensor import Tensor, empty

import math


class Linear(Module):
    def __init__(self, in_features, out_features, bias=True) -> None:
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(empty((self.in_features, self.out_features)))
        self.bias = Parameter(empty(self.out_features)) if bias else None
        self.reset_paramters()

    def reset_paramters(self):
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = init._calculate_fan(self.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            init.uniform_(self.bias, -bound, bound)

    def forward(self, x: Tensor):
        return F.linear(x, self.weight, self.bias)

    def __repr__(self) -> str:
        return "Linear(in_features={}, out_features={}, bias={})".format(
            self.in_features, self.out_features, self.bias is not None)


from .module import Module
from ...tensor import Tensor
import numpy as np


class Dropout(Module):
    def __init__(self, p: float = 0.5) -> None:
        super().__init__()
        assert p >= 0 and p < 1
        self.p = p

    def forward(self, x) -> Tensor:
        if self._train:
            return x * Tensor(np.random.binomial(1, 1 - self.p, x.shape[-1]))
        return x * (1 - self.p)

    def __repr__(self) -> str:
        return "{}(p={})".format(self.__class__.__name__, self.p)


'''优化器类，我们目前实现了\n
- SGD;\n
- Momentum;\n
- Adagrad;\n
- Adadelta;\n
- Adam.\n

Reference
---------
论文: https://arxiv.org/abs/1609.04747;\n
博客: https://welts.xyz/2021/08/20/gd/.
'''
import numpy as np


class SGD:
    '''批梯度下降优化器:

    .. math:: \\theta=\\theta-\\eta\\cdot\\nabla_{\\theta}J(\\theta;x^{(i:i+n)}, y^{(i:i+n)})

    Parameters
    ----------
    params : tuple
        待优化参数元组;
    lr : float
        学习率;
    weight_decay : float, default=0.
        权重衰减系数.
    '''
    def __init__(self,
                 params,
                 lr: float,
                 weight_decay: float = 0.) -> None:
        self.lr = lr
        self.params = list(params)
        self.weight_devay = weight_decay

    def step(self):
        '''对所有参数进行上式的更新.'''
        for param in self.params:
            param.data -= self.lr * (param.grad +
                                     self.weight_devay * param.data)

    def zero_grad(self):
        '''针对self.params梯度清零.'''
        for param in self.params:
            param.zero_grad()


class Momentum(SGD):
    '''带动量的梯度下降

    .. math::
        v_t &= \\gamma v_{t-1}+\\eta\\nabla_{\\theta}J(\\theta) \\\\
        \\theta &= \\theta-v_t

    Parameters
    ----------
    params : tuple
        待优化参数元组;
    lr : float
        学习率;
    momentum : float, default=0.5
        动量系数;
    weight_decay : float, default=0.
        权重衰减系数.
    '''
    def __init__(self,
                 params: tuple,
                 lr: float,
                 momentum: float = 0.5,
                 weight_decay: float = 0.) -> None:
        super().__init__(params, lr, weight_decay)
        self.momentum = momentum
        self.v = [np.zeros(param.shape) for param in self.params]

    def step(self):
        for i in range(len(self.params)):
            self.v[i] *= self.momentum
            self.v[i] += self.lr * (self.params[i].grad +
                                    self.weight_devay * self.params[i].data)
            self.params[i].data -= self.v[i]

    def zero_grad(self):
        return super().zero_grad()


class Adagrad(SGD):
    '''Adagrad下降法

    Parameters
    ----------
    params : tuple
        待优化参数元组;
    lr : float
        学习率;
    weight_decay : float, default=0.
        权重衰减系数.
    '''
    def __init__(self, params, lr=0.01, weight_decay=0.) -> None:
        super().__init__(params, lr, weight_decay)
        self.G = [np.zeros(param.shape) for param in self.params]

    def step(self):
        for i in range(len(self.params)):
            grad = self.params[i].grad + self.weight_devay * self.params[i].data
            self.G[i] += grad**2
            self.params[i].data -= self.lr * grad / np.sqrt(1e-8 + self.G[i])

    def zero_grad(self):
        return super().zero_grad()


class Adadelta(Adagrad):
    '''Adadelta下降法

    Parameters
    ----------
    params : tuple
        待优化参数元组;
    lr : float
        学习率;
    gamma : float, default=0.9
        滑动系数 :math:`\\gamma`，默认0.9;
    weight_decay : float, default=0.
        权重衰减系数.
    '''
    def __init__(self,
                 params: tuple,
                 lr: float,
                 gamma: float = 0.9,
                 weight_decay: float = 0.) -> None:
        super().__init__(params, lr, weight_decay)
        self.gamma = gamma

    def step(self):
        for i in range(len(self.params)):
            grad = self.params[i].grad + self.weight_devay * self.params[i].data
            self.G[i] = self.gamma * self.G[i] + (1 - self.gamma) * grad**2
            self.params[i].data -= self.lr * grad / np.sqrt(self.G[i] + 1e-8)

    def zero_grad(self):
        return super().zero_grad()


class Adam(SGD):
    '''Adam下降法
    
    Parameters
    ----------
    params : tuple
        待优化参数元组;
    lr : float
        学习率;
    beta1 : float, default=0.9
        Adam参数1，默认0.9;
    beta2 : float, default=0.999
        Adam参数2，默认0.999;
    weight_decay : float, default=0.
        权重衰减系数.
    '''
    def __init__(self,
                 params: tuple,
                 lr: float,
                 beta1: float = 0.9,
                 beta2: float = 0.999,
                 weight_decay: float = 0.) -> None:
        super().__init__(params, lr, weight_decay)
        self.beta1, self.beta2 = beta1, beta2
        self.m = [np.zeros(param.shape) for param in self.params]
        self.v = [np.zeros(param.shape) for param in self.params]
        self.t = 1

    def step(self):
        for i in range(len(self.params)):
            grad = self.params[i].grad + self.weight_devay * self.params[i].data
            self.m[i] = self.beta1 * self.m[i] + (1 - self.beta1) * grad
            self.v[i] = self.beta2 * self.v[i] + (1 - self.beta2) * grad**2
            m_t = self.m[i] / (1 - self.beta1**self.t)
            v_t = self.v[i] / (1 - self.beta2**self.t)

            self.params[i].data -= self.lr * m_t / (np.sqrt(v_t) + 1e-8)
        self.t += 1

    def zero_grad(self):
        return super().zero_grad()


from .module import Module
from ..parameter import Parameter
from .. import init
from .. import functional as F
from ... import tensor

import math


class Conv1d(Module):
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride=1,
                 padding=0,
                 bias=True) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.stride = stride
        self.weight = Parameter(
            tensor.empty(
                (self.out_channels, self.in_channels, self.kernel_size)))
        self.bias = Parameter(tensor.empty(
            (1, self.out_channels, 1))) if bias else None
        self.reset_parameters()

    def reset_parameters(self):
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = init._calculate_fan(self.weight)
            if fan_in != 0:
                bound = 1 / math.sqrt(fan_in)
                init.uniform_(self.bias, -bound, bound)

    def forward(self, x):
        conv1d = F.conv1d(x, self.weight, self.padding, self.stride)
        if self.bias is not None:
            return conv1d + self.bias
        return conv1d

    def __repr__(self) -> str:
        return "{}(in_channels={}, out_channels={}, kernel_size={}, padding={}, stride={}, bias={})".format(
            self.__class__.__name__,
            self.in_channels,
            self.out_channels,
            self.kernel_size,
            self.padding,
            self.stride,
            self.bias is not None,
        )


class Conv2d(Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        padding=0,
        stride=1,
        bias=True,
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.stride = stride
        self.weight = Parameter(
            tensor.empty((self.out_channels, self.in_channels,
                          self.kernel_size, self.kernel_size)))
        self.bias = Parameter(tensor.empty(
            (1, self.out_channels, 1, 1))) if bias else None
        self.reset_parameters()

    def reset_parameters(self):
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = init._calculate_fan(self.weight)
            if fan_in != 0:
                bound = 1 / math.sqrt(fan_in)
                init.uniform_(self.bias, -bound, bound)

    def forward(self, x):
        conv2d = F.conv2d(x, self.weight, self.padding, self.stride)
        if self.bias is not None:
            return conv2d + self.bias
        return conv2d

    def __repr__(self) -> str:
        return "{}(in_channels={}, out_channels={}, kernel_size={}, padding={}, stride={}, bias={})".format(
            self.__class__.__name__,
            self.in_channels,
            self.out_channels,
            self.kernel_size,
            self.padding,
            self.stride,
            self.bias is not None,
        )


class MaxPool1d(Module):
    def __init__(self, kernel_size, stride, padding) -> None:
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding

    def forward(self, x):
        return F.max_pool1d(x, self.kernel_size, self.stride, self.padding)

    def __repr__(self) -> str:
        return "{}(kernel_size={}, stride={}, padding={})".format(
            self.__class__.__name__,
            self.kernel_size,
            self.stride,
            self.padding,
        )


class AvgPool1d(Module):
    def __init__(self, kernel_size, stride, padding) -> None:
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding

    def forward(self, x):
        return F.avg_pool1d(x, self.kernel_size, self.stride, self.padding)

    def __repr__(self) -> str:
        return "{}(kernel_size={}, stride={}, padding={})".format(
            self.__class__.__name__,
            self.kernel_size,
            self.stride,
            self.padding,
        )


class MaxPool2d(Module):
    def __init__(self, kernel_size, stride, padding) -> None:
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding

    def forward(self, x):
        return F.max_pool2d(x, self.kernel_size, self.stride, self.padding)

    def __repr__(self) -> str:
        return "{}(kernel_size={}, stride={}, padding={})".format(
            self.__class__.__name__,
            self.kernel_size,
            self.stride,
            self.padding,
        )


class AvgPool2d(Module):
    def __init__(self, kernel_size, stride, padding) -> None:
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding

    def forward(self, x):
        return F.avg_pool2d(x, self.kernel_size, self.stride, self.padding)

    def __repr__(self) -> str:
        return "{}(kernel_size={}, stride={}, padding={})".format(
            self.__class__.__name__,
            self.kernel_size,
            self.stride,
            self.padding,
        )


from .activation import Sigmoid, Tanh, ReLU, LeakyReLU, Softmax
from .batchnorm import BatchNorm1d, BatchNorm2d
from .conv import Conv1d, Conv2d, MaxPool1d, MaxPool2d, AvgPool1d, AvgPool2d
from .dropout import Dropout
from .linear import Linear
from .loss import MSELoss, NLLLoss, CrossEntropyLoss
from .module import Module, Sequential
from .rnn import RNN, LSTM, GRU

__all__ = [
    "Sigmoid", "Tanh", "ReLU", "LeakyReLU", "Softmax", "BatchNorm1d",
    "BatchNorm2d", "Conv1d", "Conv2d", "MaxPool1d", "MaxPool2d", "AvgPool1d",
    "AvgPool2d", "Dropout", "Linear", "MSELoss", "NLLLoss", "CrossEntropyLoss",
    "Module", "Sequential", "RNN", "LSTM", "GRU"
]



import setuptools

setuptools.setup(
    name='pydynet',
    version='0.0.6',
    description=
    'Neuron network(DNN, CNN, RNN, etc) implementation using Numpy based on autodiff',
    author="Welt Xing",
    author_email="xingcy@smail.nju.edu.cn",
    maintainer="Welt Xing",
    maintainer_email="xingcy@smail.nju.edu.cn",
    packages=['pydynet', 'pydynet/nn', 'pydynet/nn/modules'],
    license='MIT License',
    install_requires=['numpy'],
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type="text/markdown",
    url='https://github.com/Kaslanarian/PyDyNet',
)



from .module import Module
from .. import functional as F
from ..parameter import Parameter
from ... import tensor


class RNN(Module):
    '''单层RNN，不像Pytorch那样可以指定num_layers进行多层堆叠，我们的RNN是单层可双向的，'''
    def __init__(
        self,
        input_size,
        hidden_size,
        nonlinearity='tanh',
        batch_first=False,
        bidirectional=False,
    ) -> None:
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.batch_first = batch_first
        self.bidirectional = bidirectional
        self.nonlinearity = nonlinearity
        self.fn = {
            "tanh": F.tanh,
            "relu": F.relu,
        }[nonlinearity]
        scale = 1 / self.hidden_size**0.5
        low_high = (-scale, scale)

        self.Wx = Parameter(
            tensor.uniform(*low_high, (input_size, hidden_size)))
        self.Wh = Parameter(
            tensor.uniform(*low_high, (hidden_size, hidden_size)))
        self.bias = Parameter(tensor.uniform(*low_high, hidden_size))

        if self.bidirectional:
            self.Wx_reverse = Parameter(
                tensor.uniform(*low_high, (input_size, hidden_size)))
            self.Wh_reverse = Parameter(
                tensor.uniform(*low_high, (hidden_size, hidden_size)))
            self.bias_reverse = Parameter(
                tensor.uniform(*low_high, hidden_size))

    def forward(self, x: tensor.Tensor, h=(None, None)):
        h, h_reverse = h
        if h is None:
            h = tensor.zeros(self.hidden_size)
        if h_reverse is None:
            h_reverse = tensor.zeros(self.hidden_size)

        if self.batch_first and x.ndim == 3:
            # x.ndim的数据是(seq_len, input_size)不需要变换
            x = x.transpose(1, 0, 2)

        h_list = []
        if self.bidirectional:
            h_reverse_list = []
            for i in range(x.shape[0]):
                h = self.fn(x[i:i + 1] @ self.Wx + h @ self.Wh + self.bias)
                h_list.append(h)
                h_reverse = self.fn(
                    x[x.shape[0] - i - 1:x.shape[0] -
                      i]) @ self.Wx + h_reverse @ self.Wh + self.bias_reverse
                h_reverse_list.append(h_reverse)
            output = tensor.concatenate(
                tensor.concatenate(*h_list),
                tensor.concatenate(*h_reverse_list),
                axis=-1,
            )
        else:
            for i in range(x.shape[0]):
                h = self.fn(x[i:i + 1] @ self.Wx + h @ self.Wh + self.bias)
                h_list.append(h)
            output = tensor.concatenate(*h_list)

        if self.batch_first and x.ndim == 3:
            output = output.transpose(1, 0, 2)
        return output

    def __repr__(self) -> str:
        return "{}(input_size={}, hidden_size={}, nonlinearity={}, batch_first={}, bidirectional={})".format(
            self.__class__.__name__, self.input_size, self.hidden_size,
            self.nonlinearity, self.batch_first, self.bidirectional)


class LSTM(Module):
    def __init__(
        self,
        input_size,
        hidden_size,
        batch_first=False,
        bidirectional=False,
    ) -> None:
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.batch_first = batch_first
        self.bidirectional = bidirectional
        scale = 1 / self.hidden_size**0.5
        low_high = (-scale, scale)

        self.Wx = Parameter(
            tensor.uniform(*low_high, (input_size, hidden_size * 4)))
        self.Wh = Parameter(
            tensor.uniform(*low_high, (hidden_size, hidden_size * 4)))
        self.bias = Parameter(tensor.uniform(*low_high, hidden_size * 4))

        if self.bidirectional:
            self.Wx_reverse = Parameter(
                tensor.uniform(*low_high, (input_size, hidden_size * 4)))
            self.Wh_reverse = Parameter(
                tensor.uniform(*low_high, (hidden_size, hidden_size * 4)))
            self.bias_reverse = Parameter(
                tensor.uniform(*low_high, hidden_size * 4))

    def forward(self, x: tensor.Tensor, h=(None, None)) -> tensor.Tensor:
        c = tensor.zeros(self.hidden_size)
        c_reverse = tensor.zeros(self.hidden_size)

        h, h_reverse = h
        if h is None:
            h = tensor.zeros(self.hidden_size)
        if h_reverse is None:
            h_reverse = tensor.zeros(self.hidden_size)

        if self.batch_first and x.ndim == 3:
            x = x.transpose(1, 0, 2)

        h_list = []

        if self.bidirectional:
            h_reverse_list = []
            for id in range(x.shape[0]):
                affine = x[id:id + 1] @ self.Wx + h @ self.Wh + self.bias
                f_i_o = affine[..., :3 * self.hidden_size]
                g = affine[..., -self.hidden_size:]
                sigma_fio = F.sigmoid(f_i_o)
                g = F.tanh(g)
                f = sigma_fio[..., :self.hidden_size]
                i = sigma_fio[..., self.hidden_size:2 * self.hidden_size]
                o = sigma_fio[..., -self.hidden_size:]
                c = tensor.sum(f * c + g * i, (0, 1))
                h = o * F.tanh(c)
                h_list.append(h)

                affine = x[
                    x.shape[0] - id - 1:x.shape[0] -
                    id] @ self.Wx_reverse + h_reverse @ self.Wh_reverse + self.bias_reverse
                f_i_o = affine[..., :3 * self.hidden_size]
                g = affine[..., -self.hidden_size:]
                sigma_fio = F.sigmoid(f_i_o)
                g = F.tanh(g)
                f = sigma_fio[..., :self.hidden_size]
                i = sigma_fio[..., self.hidden_size:2 * self.hidden_size]
                o = sigma_fio[..., -self.hidden_size:]
                c = tensor.sum(f * c_reverse + g * i, (0, 1))
                h_reverse = o * F.tanh(c_reverse)
                h_reverse_list.append(h_reverse)

                output = tensor.concatenate(
                    tensor.concatenate(*h_list),
                    tensor.concatenate(*h_reverse_list),
                    axis=-1,
                )
        else:
            for id in range(x.shape[0]):
                affine = x[id:id + 1] @ self.Wx + h @ self.Wh + self.bias
                f_i_o = affine[..., :3 * self.hidden_size]
                g = affine[..., -self.hidden_size:]
                sigma_fio = F.sigmoid(f_i_o)
                g = F.tanh(g)
                f = sigma_fio[..., :self.hidden_size]
                i = sigma_fio[..., self.hidden_size:2 * self.hidden_size]
                o = sigma_fio[..., -self.hidden_size:]
                c = F.mean(f * c + g * i, (0, 1))
                h = o * F.tanh(c)
                h_list.append(h)
            output = tensor.concatenate(*h_list)

        if self.batch_first and x.ndim == 3:
            output = output.transpose(1, 0, 2)
        return output

    def __repr__(self) -> str:
        return "{}(input_size={}, hidden_size={}, batch_first={}, bidirectional={})".format(
            self.__class__.__name__, self.input_size, self.hidden_size,
            self.batch_first, self.bidirectional)


class GRU(Module):
    def __init__(
        self,
        input_size,
        hidden_size,
        batch_first=False,
        bidirectional=False,
    ) -> None:
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.batch_first = batch_first
        self.bidirectional = bidirectional
        scale = 1 / self.hidden_size**0.5
        low_high = (scale, -scale)

        self.Wx = Parameter(
            tensor.uniform(*low_high, (input_size, 3 * hidden_size)))
        self.Wh_zr = Parameter(
            tensor.uniform(*low_high, (hidden_size, 2 * hidden_size)))
        self.bias_zr = Parameter(tensor.uniform(*low_high, 2 * hidden_size))
        self.Wh = Parameter(
            tensor.uniform(*low_high, (hidden_size, hidden_size)))
        self.bias = Parameter(tensor.uniform(*low_high, self.hidden_size))

        if self.bidirectional:
            self.Wx_reverse = Parameter(
                tensor.uniform(*low_high, (input_size, 3 * hidden_size)))
            self.Wh_zr_reverse = Parameter(
                tensor.uniform(*low_high, (hidden_size, 2 * hidden_size)))
            self.bias_zr_recerse = Parameter(
                tensor.uniform(*low_high, 2 * hidden_size))
            self.Wh_reverse = Parameter(
                tensor.uniform(*low_high, (hidden_size, hidden_size)))
            self.bias_reverse = Parameter(
                tensor.uniform(*low_high, self.hidden_size))

    def forward(self, x: tensor.Tensor, h=(None, None)) -> tensor.Tensor:
        h, h_reverse = h
        if h is None:
            h = tensor.zeros(self.hidden_size)
        if h_reverse is None:
            h_reverse = tensor.zeros(self.hidden_size)

        if self.batch_first and x.ndim == 3:
            # x.ndim的数据是(seq_len, input_size)不需要变换
            x = x.transpose(1, 0, 2)

        h_list = []

        if self.bidirectional:
            h_reverse_list = []
            for i in range(x.shape[0]):
                affine = x[i:i + 1] @ self.Wx
                zr = F.sigmoid(affine[..., :2 * self.hidden_size] +
                               h @ self.Wh_zr + self.bias_zr)
                z, r = zr[..., :self.hidden_size], zr[..., self.hidden_size:]
                h_tilde = F.tanh(affine[..., -self.hidden_size:] +
                                 (r * h) @ self.Wh + self.bias)
                h = (1 - z) * h + z * h_tilde
                h_list.append(h)

                affine = x[x.shape[0] - i - 1:x.shape[0] - i] @ self.Wx_reverse
                zr = F.sigmoid(affine[..., :2 * self.hidden_size] +
                               h @ self.Wh_zr_reverse + self.bias_zr_recerse)
                z, r = zr[..., :self.hidden_size], zr[..., self.hidden_size:]
                h_tilde = F.tanh(affine[..., -self.hidden_size:] +
                                 (r * h) @ self.Wh_reverse + self.bias_reverse)
                h_reverse = (1 - z) * h_reverse + z * h_tilde
                h_reverse_list.append(h_reverse)

                output = tensor.concatenate(
                    tensor.concatenate(*h_list),
                    tensor.concatenate(*h_reverse_list),
                    axis=-1,
                )

        else:
            for i in range(x.shape[0]):
                affine = x[i:i + 1] @ self.Wx
                zr = F.sigmoid(affine[..., :2 * self.hidden_size] +
                               h @ self.Wh_zr + self.bias_zr)
                z, r = zr[..., :self.hidden_size], zr[..., self.hidden_size:]
                h_tilde = F.tanh(affine[..., -self.hidden_size:] +
                                 (r * h) @ self.Wh + self.bias)
                h = (1 - z) * h + z * h_tilde
                h_list.append(h)
            output = tensor.concatenate(*h_list)

        if self.batch_first and x.ndim == 3:
            output = output.transpose(1, 0, 2)
        return output

    def __repr__(self) -> str:
        return "{}(input_size={}, hidden_size={}, batch_first={}, bidirectional={})".format(
            self.__class__.__name__, self.input_size, self.hidden_size,
            self.batch_first, self.bidirectional)


import numpy as np
from .autograd import is_grad_enable, no_grad


class Graph:
    '''计算图，全局共用一个动态计算图'''
    node_list: list = list()

    @classmethod
    def add_node(cls, node):
        '''添加静态图节点'''
        cls.node_list.append(node)

    @classmethod
    def clear(cls):
        '''清空计算图'''
        cls.node_list.clear()

    @classmethod
    def free_graph(cls):
        '''
        释放计算图，和clear的区别在于我们不会删除叶子节点，
        这一点和PyTorch类似。
        '''
        new_list = []
        for node in Graph.node_list:
            node.next.clear()
            if node.is_leaf:
                # 叶子节点
                new_list.append(node)

            node.last.clear()
        Graph.node_list = new_list


class Tensor:
    '''
    将数据(NumPy数组)包装成可微分张量

    Parameters
    ----------
    data : ndarray
        张量数据，只要是np.array能够转换的数据;
    requires_grad : bool, default=False
        是否需要求梯度;
    dtype : default=None
        数据类型，和numpy数组的dtype等价

    Attributes
    ----------
    data : numpy.ndarray
        核心数据，为NumPy数组;
    requires_grad : bool
        是否需要求梯度;
    grad : numpy.ndarray
        梯度数据，为和data相同形状的数组(初始化为全0);
    next : list[Tensor]
        下游节点列表；
    last : list[Tensor]
        上游节点列表.

    Example
    -------
    >>> import numpy as np
    >>> from pydynet.tensor import Tensor
    >>> x = Tensor(1., requires_grad=True)
    >>> y = Tensor([1, 2, 3], dtype=float)
    >>> z = Tensor(np.random.rand(3, 4))
    '''
    def __init__(self, data, requires_grad: bool = False, dtype=None) -> None:
        if isinstance(data, Tensor):
            data = data.data
        self.data: np.ndarray = np.array(data, dtype)
        self.requires_grad: bool = requires_grad and is_grad_enable()
        assert not (
            self.requires_grad and self.dtype != float
            and self.dtype != complex
        ), "Only Tensors of floating point and complex dtype can require gradients"
        self.grad: np.ndarray = np.zeros_like(
            self.data) if self.requires_grad else None
        self.next: list = list()
        self.last: list = list()
        if self.requires_grad:
            # 不需要求梯度的节点不出现在动态计算图中
            Graph.add_node(self)

    @property
    def is_leaf(self):
        '''判断是否为叶节点:需要求导且无上游节点的节点为叶节点.'''
        return not self.requires_grad or len(self.last) == 0

    @property
    def shape(self):
        '''张量的形状，用法同NumPy.
        
        Example
        -------
        >>> from pydynet import Tensor
        >>> Tensor([[2, 2]]).shape
        (1, 2)
        '''
        return self.data.shape

    @property
    def ndim(self):
        '''张量的维度，用法同NumPy.
        
        Example
        -------
        >>> from pydynet import Tensor
        >>> Tensor([[2, 2]]).ndim
        2
        '''
        return self.data.ndim

    @property
    def dtype(self):
        '''张量的数据类型，用法同NumPy.

        Example
        -------
        >>> from pydynet import Tensor
        >>> Tensor([[2, 2]]).dtype
        dtype('int64')
        '''
        return self.data.dtype

    @property
    def size(self):
        '''张量的元素个数，用法同NumPy.

        Example
        -------
        >>> from pydynet import Tensor
        >>> Tensor([[1, 1]]).size
        2
        '''
        return self.data.size

    @property
    def T(self):
        return self.transpose()

    def astype(self, new_type):
        '''类型转换，我们不允许可求导节点的类型转换'''
        assert self.requires_grad
        self.data.astype(new_type)

    def reshape(self, *new_shape):
        return reshape(self, new_shape)

    def transpose(self, *axes):
        return transpose(self, axes if len(axes) != 0 else None)

    def max(self, axis=None, keepdims=False):
        return max(self, axis, keepdims)

    def min(self, axis=None, keepdims=False):
        return min(self, axis, keepdims)

    def mean(self, axis=None, keepdims=False):
        return mean(self, axis, keepdims)

    def sum(self, axis=None, keepdims=False):
        return sum(self, axis, keepdims)

    def argmax(self, axis=None):
        return argmax(self, axis)

    def argmin(self, axis=None):
        return argmin(self, axis)

    def build_edge(self, node):
        '''构建两节点的有向边，正常不适用'''
        self.next.append(node)
        node.last.append(self)

    def __repr__(self) -> str:
        return "{}({}, requires_grad={})".format(
            "tensor",
            self.data,
            self.requires_grad,
        )

    def __add__(self, x):
        return add(self, x)

    def __radd__(self, x):
        return add(x, self)

    def __sub__(self, x):
        return sub(self, x)

    def __rsub__(self, x):
        return sub(x, self)

    def __mul__(self, x):
        return mul(self, x)

    def __rmul__(self, x):
        return mul(x, self)

    def __matmul__(self, x):
        return matmul(self, x)

    def __rmatmul__(self, x):
        return matmul(x, self)

    def __truediv__(self, x):
        return div(self, x)

    def __rtruediv__(self, x):
        return div(x, self)

    def __pow__(self, x):
        return pow(self, x)

    def __rpow__(self, x):
        return pow(x, self)

    def __pos__(self):
        return 1 * self

    def __neg__(self):
        return -1 * self

    def __abs__(self):
        return abs(self)

    def __getitem__(self, key):
        return get_slice(self, key)

    def __setitem__(self, key, value):
        '''
        重载了切片/索引赋值的操作，我们不允许self允许求导，否则将出现错误

        Parameters
        ----------
        key : 索引，支持NumPy的数字、切片和条件索引
        value : 值，可以是NumPy数字，也可以是数字

        Example
        -------
        >>> x = Tensor([1, 2, 3])
        >>> x[x <= 2] = 0
        >>> x
        <[0 0 3], int64, Tensor>
        '''
        assert not self.requires_grad, "In-place operation is forbidden in node requires grad."
        if isinstance(key, Tensor):
            key = key.data
        if not isinstance(value, Tensor):
            self.data[key] = value
        else:
            self.data[key] = value.data

    def __len__(self):
        return len(self.data)

    def __iadd__(self, other):
        assert not self.requires_grad, "In-place operation is forbidden in node requires grad."
        if isinstance(other, Tensor):
            other = other.data
        self.data += other
        return self

    def __isub__(self, other):
        assert not self.requires_grad, "In-place operation is forbidden in node requires grad."
        if isinstance(other, Tensor):
            other = other.data
        self.data -= other
        return self

    def __imul__(self, other):
        assert not self.requires_grad, "In-place operation is forbidden in node requires grad."
        if isinstance(other, Tensor):
            other = other.data
        self.data *= other
        return self

    def __itruediv__(self, other):
        assert not self.requires_grad, "In-place operation is forbidden in node requires grad."
        if isinstance(other, Tensor):
            other = other.data
        self.data /= other
        return self

    def __imatmul__(self, other):
        assert not self.requires_grad, "In-place operation is forbidden in node requires grad."
        if isinstance(other, Tensor):
            other = other.data
        self.data @= other
        return self

    @no_grad()
    def __lt__(self, other):
        if isinstance(other, Tensor):
            other = other.data
        return Tensor(self.data < other)

    @no_grad()
    def __le__(self, other):
        if isinstance(other, Tensor):
            other = other.data
        return Tensor(self.data <= other)

    # 这里没有重载__eq__和__neq__是因为在RNN中这样的重载会引发问题
    @no_grad()
    def eq(self, other):
        if isinstance(other, Tensor):
            other = other.data
        return Tensor(self.data == other)

    @no_grad()
    def ne(self, other):
        if isinstance(other, Tensor):
            other = other.data
        return Tensor(self.data != other)

    @no_grad()
    def __gt__(self, other):
        if isinstance(other, Tensor):
            other = other.data
        return Tensor(self.data > other)

    @no_grad()
    def __ge__(self, other):
        if isinstance(other, Tensor):
            other = other.data
        return Tensor(self.data >= other)

    def backward(self, retain_graph=False):
        '''
        以节点为输出进行反向传播

        Parameters
        ----------
        retain_graph : bool, default=False
            是否保留计算图

        Example
        -------
        >>> from pydynet.tensor import Tensor
        >>> import pydynet.functional as F
        >>> x = Tensor(2., requires_grad=True)
        >>> y = x**2 + x - 1
        >>> y.backward()
        >>> x.grad
        5.
        '''
        if self not in Graph.node_list:
            print("AD failed because the node is not in graph.")
            return

        assert self.data.ndim == 0, "backward should be called only on a scalar."

        self.grad = np.ones_like(self.data)
        for i in range(len(Graph.node_list) - 1, -1, -1):
            if Graph.node_list[i] is self:
                y_id = i
                break

        for node in Graph.node_list[y_id::-1]:
            grad = node.grad
            for last in [l for l in node.last if l.requires_grad]:
                add_grad = node.grad_fn(last, grad)
                # 广播机制处理梯度
                if add_grad.shape != last.shape:
                    add_grad = np.sum(
                        add_grad,
                        axis=tuple(-i for i in range(1, last.ndim + 1)
                                   if last.shape[-i] == 1),
                        keepdims=True,
                    )
                    add_grad = np.sum(
                        add_grad,
                        axis=tuple(range(add_grad.ndim - last.ndim)),
                    )
                last.grad += add_grad

            if not node.is_leaf:
                node.grad = None

        if not retain_graph:
            Graph.free_graph()

    def zero_grad(self):
        '''梯度归零'''
        self.grad = np.zeros(self.shape)

    def numpy(self) -> np.ndarray:
        '''返回Tensor的内部数据，即NumPy数组(拷贝)'''
        return self.data.copy()

    def item(self):
        return self.data.item()


class UnaryOperator(Tensor):
    '''
    一元运算算子的基类，将一个一元函数抽象成类

    Example
    -------
    >>> class exp(UnaryOperator):
            def forward(self, x: Tensor):
                return np.exp(x.data)
            def grad_fn(self, x: Tensor, grad) -> np.ndarray:
                return self.data * grad
    >>> x = Tensor(1.)
    >>> y = exp(x)
    '''
    def __init__(self, x: Tensor) -> None:
        if not isinstance(x, Tensor):
            x = Tensor(x)
        super().__init__(
            self.forward(x),
            is_grad_enable() and x.requires_grad,
        )
        if self.requires_grad:
            x.build_edge(self)

    def forward(self, x: Tensor) -> np.ndarray:
        '''前向传播函数，参数为Tensor，返回的是NumPy数组'''
        raise NotImplementedError

    def grad_fn(self, x: Tensor, grad: np.ndarray) -> np.ndarray:
        '''
        反向传播函数，参数为下游节点，从上游流入该节点梯度。
        注："上游"和"下游"针对的是反向传播，比如z = f(x, y)，x和y是z的下游节点.

        x : Tensor
            下游节点
        grad : ndarray
            上游流入该节点的梯度
        '''
        raise NotImplementedError

    def __repr__(self) -> str:
        return "Tensor({}, op={})".format(self.data, self.__class__.__name__)


class BinaryOperator(Tensor):
    '''
    二元运算算子的基类，将一个二元函数抽象成类

    Example
    -------
    >>> add(BinaryOperator):
            def forward(self, x: Tensor, y: Tensor):
                return x.data + y.data
            def grad_fn(self, node: Tensor, grad: np.ndarray):
                return np.ones(self.shape) * grad
    >>> x = Tensor(1.)
    >>> y = Tensor(2.)
    >>> z = add(x, y)
    '''
    def __init__(self, x: Tensor, y: Tensor) -> None:
        if not isinstance(x, Tensor):
            x = Tensor(x)
        if not isinstance(y, Tensor):
            y = Tensor(y)
        super().__init__(
            self.forward(x, y),
            is_grad_enable() and (x.requires_grad or y.requires_grad),
        )
        if self.requires_grad:
            x.build_edge(self)
            y.build_edge(self)

    def forward(self, x: Tensor, y: Tensor) -> np.ndarray:
        '''前向传播函数，参数为Tensor，返回的是NumPy数组'''
        raise NotImplementedError

    def grad_fn(self, x: Tensor, grad) -> np.ndarray:
        '''
        反向传播函数，参数为下游节点，从上游流入该节点梯度。
        注："上游"和"下游"针对的是反向传播，比如z = f(x, y)，x和y是z的下游节点.

        x : Tensor
            下游节点
        grad : ndarray
            上游流入该节点的梯度
        '''
        raise NotImplementedError

    def __repr__(self) -> str:
        return "Tensor({}, op={})".format(self.data, self.__class__.__name__)


class MultiOperator(Tensor):
    '''多元运算算子基类'''
    def __init__(self, *tensors) -> None:
        requires_grad = False
        tensors = list(tensors)
        if is_grad_enable():
            for i in range(len(tensors)):
                if not isinstance(tensors[i], Tensor):
                    tensors[i] = Tensor(tensors[i])
                requires_grad = requires_grad or tensors[i].requires_grad

        super().__init__(self.forward(*tensors), requires_grad=requires_grad)
        if self.requires_grad:
            for i in range(len(tensors)):
                tensors[i].build_edge(self)

    def forward(self, *tensors) -> np.ndarray:
        raise NotImplementedError

    def grad_fn(self, x: Tensor, grad: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def __repr__(self) -> str:
        return "Tensor({}, op={})".format(self.data, self.__class__.__name__)


class add(BinaryOperator):
    '''
    加法算子

    Example
    -------
    >>> x = Tensor(1.)
    >>> y = Tensor(2.)
    >>> z = add(x, y) # 在Tensor类中进行了重载，所以也可以写成
    >>> z = x + y
    '''
    def forward(self, x: Tensor, y: Tensor):
        return x.data + y.data

    def grad_fn(self, node: Tensor, grad: np.ndarray):
        return grad[...]


class sub(BinaryOperator):
    '''
    减法算子，在Tensor类中进行重载

    See also
    --------
    add : 加法算子
    '''
    def forward(self, x: Tensor, y: Tensor):
        return x.data - y.data

    def grad_fn(self, node: Tensor, grad: np.ndarray):
        if node is self.last[0]:
            return grad[...]
        return -grad


class mul(BinaryOperator):
    '''
    元素级乘法算子，在Tensor类中进行重载

    Example
    -------
    >>> x = Tensor([1., 2.])
    >>> y = Tensor([2., 3.])
    >>> z = mul(x, y) # [2, 6]

    See also
    --------
    add : 加法算子
    '''
    def __init__(self, x: Tensor, y: Tensor) -> None:
        super().__init__(x, y)

    def forward(self, x: Tensor, y: Tensor):
        return x.data * y.data

    def grad_fn(self, node: Tensor, grad: np.ndarray):
        if node is self.last[0]:
            return self.last[1].data * grad
        return self.last[0].data * grad


class div(BinaryOperator):
    '''
    除法算子，在Tensor类中进行重载

    See also
    --------
    add : 加法算子
    '''
    def __init__(self, x: Tensor, y: Tensor) -> None:
        super().__init__(x, y)

    def forward(self, x: Tensor, y: Tensor):
        return x.data / y.data

    def grad_fn(self, node: Tensor, grad: np.ndarray):
        temp = grad / self.last[1].data
        if node is self.last[0]:
            return temp
        return -self.data * temp


class pow(BinaryOperator):
    '''
    幂运算算子，在Tensor类中进行重载

    See also
    --------
    add : 加法算子
    '''
    def __init__(self, x: Tensor, y: Tensor) -> None:
        super().__init__(x, y)

    def forward(self, x: Tensor, y: Tensor):
        return x.data**y.data

    def grad_fn(self, node: Tensor, grad: np.ndarray) -> np.ndarray:
        if node is self.last[0]:
            return (self.data * self.last[1].data / node.data) * grad
        else:
            return self.data * np.log(self.last[0].data) * grad


class matmul(BinaryOperator):
    '''
    矩阵乘法算子，在Tensor类中进行重载，张量的矩阵乘法遵从NumPy Matmul的规则.

    参考 : https://welts.xyz/2022/04/26/broadcast/

    See also
    --------
    add : 加法算子
    '''
    def __init__(self, x: Tensor, y: Tensor) -> None:
        super().__init__(x, y)

    def forward(self, x: Tensor, y: Tensor) -> np.ndarray:
        return x.data @ y.data

    def grad_fn(self, node: Tensor, grad: np.ndarray) -> np.ndarray:
        if node is self.last[0]:
            if self.last[1].ndim == 1:
                return np.expand_dims(grad, -1) @ np.expand_dims(
                    self.last[1].data, -2)
            elif self.last[1].ndim > 2:
                shape = list(range(self.last[1].ndim))
                shape[-1], shape[-2] = shape[-2], shape[-1]
                return grad @ self.last[1].data.transpose(*shape)
            return grad @ self.last[1].data.T
        else:
            if self.last[0].ndim == 1:
                return np.expand_dims(self.last[0].data, -1) @ np.expand_dims(
                    grad, -2)
            elif self.last[0].ndim > 2:
                shape = list(range(self.last[0].ndim))
                shape[-1], shape[-2] = shape[-2], shape[-1]
                return self.last[0].data.transpose(*shape) @ grad
            return self.last[0].data.T @ grad


class abs(UnaryOperator):
    '''
    绝对值算子，在Tensor类中进行重载

    See also
    --------
    add : 加法算子
    '''
    def forward(self, x: Tensor) -> np.ndarray:
        return np.abs(x)

    def grad_fn(self, x: Tensor, grad: np.ndarray) -> np.ndarray:
        mask = np.zeros(x.shape)
        mask[x > 0] = 1.
        mask[x < 0] = -1.
        return grad * mask


class sum(UnaryOperator):
    '''
    求和算子，在Tensor类中扩展为类方法

    Parameters
    ----------
    axis : None
        求和方向(轴)
    keepdims : bool, default=False
        是否保留原来维度

    Example
    -------
    >>> x = Tensor(
            [[1, 2, 3],
            [4, 5, 6]]
        )
    >>> s1 = x.sum(0) # [5, 7, 9]
    >>> s2 = x.sum(1) # [6, 15]
    >>> s3 = sum(x, keepdims=True) # [[21]]
    ```
    '''
    def __init__(self, x: Tensor, axis=None, keepdims=False) -> None:
        self.axis = axis
        self.keepdims = keepdims
        super().__init__(x)

    def forward(self, x: Tensor) -> np.ndarray:
        return np.sum(x.data, axis=self.axis, keepdims=self.keepdims)

    def grad_fn(self, x: Tensor, grad: np.ndarray) -> np.ndarray:
        if not (self.axis is None or self.keepdims):
            grad = np.expand_dims(grad, axis=self.axis)
        return np.ones(x.shape) * grad


class mean(UnaryOperator):
    '''
    求均值算子，在Tensor类中扩展为类方法

    Parameters
    ----------
    axis : None
        求均值方向(轴)
    keepdims : bool, default=False
        是否保留原来维度

    See also
    --------
    sum : 求和算子
    '''
    def __init__(self, x: Tensor, axis=None, keepdims=False) -> None:
        self.axis = axis
        self.keepdims = keepdims
        super().__init__(x)

    def forward(self, x: Tensor) -> np.ndarray:
        return np.mean(x.data, axis=self.axis, keepdims=self.keepdims)

    def grad_fn(self, x: Tensor, grad: np.ndarray) -> np.ndarray:
        if not (self.axis is None or self.keepdims):
            grad = np.expand_dims(grad, axis=self.axis)
        return np.ones(x.shape) * grad * self.data.size / x.data.size


class max(UnaryOperator):
    '''
    求最大值算子，在Tensor类中扩展为类方法

    Parameters
    ----------
    axis : None
        求最大值方向(轴)
    keepdims : bool, default=False
        是否保留原来维度

    See also
    --------
    sum : 求和算子
    '''
    def __init__(self, x: Tensor, axis=None, keepdims=False) -> None:
        self.axis = axis
        self.keepdims = keepdims
        super().__init__(x)

    def forward(self, x: Tensor) -> np.ndarray:
        return np.max(x.data, axis=self.axis, keepdims=self.keepdims)

    def grad_fn(self, x: Tensor, grad: np.ndarray) -> np.ndarray:
        if self.keepdims or self.axis is None:
            full_dim_y = self.data
        else:
            # 还原维度
            full_dim_y = np.expand_dims(self.data, axis=self.axis)
            grad = np.expand_dims(grad, axis=self.axis)
        return (full_dim_y == x.data).astype(float) * grad


class min(UnaryOperator):
    '''
    求最小值算子，在Tensor类中扩展为类方法

    Parameters
    ----------
    axis : None
        求最大值方向(轴)
    keepdims : bool, default=False
        是否保留原来维度

    See also
    --------
    max : 求最大值算子
    '''
    def __init__(self, x: Tensor, axis=None, keepdims=False) -> None:
        self.axis = axis
        self.keepdims = keepdims
        super().__init__(x)

    def forward(self, x: Tensor) -> np.ndarray:
        return np.min(x.data, axis=self.axis, keepdims=self.keepdims)

    def grad_fn(self, x: Tensor, grad: np.ndarray) -> np.ndarray:
        if self.keepdims or self.axis is None:
            full_dim_y = self.data
        else:
            # 还原维度
            full_dim_y = np.expand_dims(self.data, axis=self.axis)
            grad = np.expand_dims(grad, axis=self.axis)
        return (full_dim_y == x.data).astype(float) * grad


class argmax(Tensor):
    def __init__(self, x: Tensor, axis=None) -> None:
        if not isinstance(x, Tensor):
            x = Tensor(x)
        self.axis = axis
        super().__init__(self.forward(x))

    def forward(self, x: Tensor) -> np.ndarray:
        return np.argmax(x.data, axis=self.axis)


class argmin(Tensor):
    def __init__(self, x: Tensor, axis=None) -> None:
        if not isinstance(x, Tensor):
            x = Tensor(x)
        self.axis = axis
        super().__init__(self.forward(x))

    def forward(self, x: Tensor) -> np.ndarray:
        return np.argmin(x.data, axis=self.axis)


class exp(UnaryOperator):
    '''指数运算
    
    Example
    -------
    >>> x = Tensor(1.)
    >>> y = exp(x)
    '''
    def forward(self, x: Tensor):
        return np.exp(x.data)

    def grad_fn(self, x: Tensor, grad: np.ndarray) -> np.ndarray:
        return self.data * grad


class log(UnaryOperator):
    '''对数运算
    
    Example
    -------
    >>> x = Tensor(1.)
    >>> y = log(x)
    '''
    def forward(self, x: Tensor):
        return np.log(x.data)

    def grad_fn(self, x: Tensor, grad: np.ndarray) -> np.ndarray:
        return grad / x.data


class maximum(BinaryOperator):
    def forward(self, x: Tensor, y: Tensor) -> np.ndarray:
        return np.maximum(x.data, y.data)

    def grad_fn(self, x: Tensor, grad) -> np.ndarray:
        return (self.data == x.data) * grad


class minimum(BinaryOperator):
    def forward(self, x: Tensor, y: Tensor) -> np.ndarray:
        return np.minimum(x, y)

    def grad_fn(self, x: Tensor, grad) -> np.ndarray:
        return (self.data == x) * grad


def sqrt(x: Tensor):
    '''平方根函数'''
    return x**0.5


def square(x: Tensor):
    '''平方函数'''
    return x * x


# 非计算函数
class reshape(UnaryOperator):
    '''
    张量形状变换算子，在Tensor中进行重载

    Parameters
    ----------
    new_shape : tuple
        变换后的形状，用法同NumPy
    '''
    def __init__(self, x: Tensor, new_shape: tuple) -> None:
        self.new_shape = new_shape
        super().__init__(x)

    def forward(self, x: Tensor) -> np.ndarray:
        return x.data.reshape(self.new_shape)

    def grad_fn(self, x: Tensor, grad: np.ndarray) -> np.ndarray:
        return grad.reshape(x.shape)


class transpose(UnaryOperator):
    '''
    张量转置算子，在Tensor中进行重载(Tensor.T和Tensor.transpose)

    Parameters
    ----------
    axes : tuple
        转置的轴变换，用法同NumPy
    '''
    def __init__(self, x: Tensor, axes: tuple = None) -> None:
        self.axes = axes
        super().__init__(x)

    def forward(self, x: Tensor) -> np.ndarray:
        return x.data.transpose(self.axes)

    def grad_fn(self, x: Tensor, grad: np.ndarray) -> np.ndarray:
        if self.axes is None:
            return grad.transpose()
        return grad.transpose(np.argsort(self.axes))


class get_slice(UnaryOperator):
    '''
    切片算子，为Tensor类提供索引和切片接口

    Example
    -------
    >>> x = Tensor(
            np.arange(12).reshape(3, 4).astype(float),
            requires_grad=True,
        )
    >>> y = x[:2, :2].sum()
    >>> y.backward()
    >>> x.grad 
    [[1. 1. 0. 0.]
     [1. 1. 0. 0.]
     [0. 0. 0. 0.]]
    '''
    def __init__(self, x: Tensor, key) -> None:
        if isinstance(key, Tensor):
            self.key = key.data
        else:
            self.key = key
        super().__init__(x)

    def forward(self, x: Tensor) -> np.ndarray:
        return x.data[self.key]

    def grad_fn(self, x: Tensor, grad: np.ndarray) -> np.ndarray:
        full_grad = np.zeros(x.shape)
        full_grad[self.key] = grad
        return full_grad


class concatenate(Tensor):
    '''对多个张量进行连接，用法类似于`numpy.concatenate`
    
    Parameters
    ----------
    *tensors : 
        待连接的张量：
    axis : default=0
        连接轴，默认是沿着第一个轴拼接.
    '''
    def __init__(self, *tensors, axis=0) -> None:
        requires_grad = False
        self.tensors = list(tensors)
        self.axis = axis
        self.indices = [0]
        for i in range(len(self.tensors)):
            if not isinstance(tensors[i], Tensor):
                self.tensors[i] = Tensor(tensors[i])
            requires_grad = requires_grad or self.tensors[i].requires_grad
            self.indices.append(self.indices[-1] +
                                self.tensors[i].shape[self.axis])

        super().__init__(self.forward(), requires_grad=requires_grad)
        if self.requires_grad:
            for i in range(len(self.tensors)):
                self.tensors[i].build_edge(self)

    def forward(self):
        return np.concatenate([t.data for t in self.tensors], axis=self.axis)

    def grad_fn(self, x, grad: np.ndarray):
        x_id = self.tensors.index(x)
        start = self.indices[x_id]
        end = self.indices[x_id + 1]
        slc = [slice(None)] * len(grad.shape)
        slc[self.axis] = slice(start, end)
        return grad[tuple(slc)]


# 一些包装的特殊矩阵
def zeros(shape, requires_grad=False):
    '''全零张量
    
    Parameters
    ----------
    shape : 
        张量形状
    require_grad : bool, default=False
        是否需要求导
    '''
    return Tensor(np.zeros(shape), requires_grad=requires_grad)


def ones(shape, requires_grad=False):
    '''全1张量
    
    Parameters
    ----------
    shape : 
        张量形状
    require_grad : bool, default=False
        是否需要求导
    '''
    return Tensor(np.ones(shape), requires_grad=requires_grad)


def randn(*shape, requires_grad=False):
    '''0-1正态分布张量
    
    Parameters
    ----------
    *shape : 
        张量形状
    require_grad : bool, default=False
        是否需要求导
    '''
    return Tensor(np.random.randn(*shape), requires_grad=requires_grad)


def rand(*shape, requires_grad=False):
    '''[0, 1)均匀分布张量
    
    Parameters
    ----------
    *shape : 
        张量形状
    require_grad : bool, default=False
        是否需要求导
    '''
    return Tensor(np.random.rand(*shape), requires_grad=requires_grad)


def uniform(low: float, high: float, shape=None, requires_grad=False):
    '''均匀分布张量
    
    Parameters
    ----------
    low : float
        均匀分布下界;
    high : float
        均匀分布下界;
    *shape : 
        张量形状
    require_grad : bool, default=False
        是否需要求导
    '''
    return Tensor(np.random.uniform(low, high, size=shape),
                  requires_grad=requires_grad)


def empty(shape, requires_grad=False):
    return Tensor(np.empty(shape), requires_grad=requires_grad)


from ..tensor import Tensor


class Parameter(Tensor):
    def __init__(self, data: Tensor) -> None:
        super().__init__(data.data, True, float)

    def __repr__(self) -> str:
        return "Parameter : {}".format(self.data)